#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "device.h"
#include "serial.h"

static int init(void);
static void destroy(void);
static int detect(void);
static int start(void);
static void stop(void);
static int pending(void);
static int getevent(device_event *ev);

static int opendev(const char *dev);
static int proc_packets(void);
static void enqueue_event(device_event *ev);
static int parse_motion(int *motv, const char *data);
static int parse_keystate(unsigned int *st, const char *data);


static int devfd = -1;
static int bnstate;

static device_event evq[16];
static int evq_widx, evq_ridx;
#define QNEXT(x)	(((x) + 1) & 0xf)
#define QEMPTY(b)	(b##_ridx == b##_widx)


struct device *device_magellan(void)
{
	static struct device dev = {
		"magellan",
		0,
		init,
		destroy,
		detect,
		start,
		stop,
		pending,
		getevent
	};
	return &dev;
}


static int init(void)
{
	return 0;
}

static void destroy(void)
{
}

static int opendev(const char *dev)
{
	int fd, timeout;
	char buf[128];

	if((fd = ser_open(dev, 9600, SER_8N2 | SER_HWFLOW)) == -1) {
		return -1;
	}

	/* try for about 5 sec */
	timeout = 5000;
	do {
		ser_printf(fd, "z\r");
	} while(!ser_wait(fd, 250) && (timeout -= 250) > 0);

	if(timeout <= 0) {
		fprintf(stderr, "magellan open(%s): device does not respond\n", dev);
		ser_close(fd);
		return -1;	/* failed to get a response */
	}

	ser_block(fd);

	ser_printf(fd, "vQ\r");
	do {
		ser_getline(fd, buf, sizeof buf);
	} while(buf[0] != 'v');

	if(buf[0] != 'v' || !strstr(buf, "MAGELLAN")) {
		fprintf(stderr, "unknown device: \"%s\"\n", buf + 1);
		ser_close(fd);
		return -1;
	}

	ser_nonblock(fd);
	return fd;
}

static int detect(void)
{
	int fd;
	if((fd = opendev(get_port())) >= 0) {
		ser_close(fd);
		return 1;
	}
	return 0;
}

static int start(void)
{
	if(devfd >= 0) {
		fprintf(stderr, "magellan start: already started\n");
		return -1;
	}

	if((devfd = opendev(get_port())) < 0) {
		return -1;
	}
	ser_printf(devfd, "m3\r");	/* start motion packets */

	return devfd;
}

static void stop(void)
{
	if(devfd < 0) {
		fprintf(stderr, "magellan stop: not started\n");
		return;
	}

	ser_close(devfd);
	devfd = -1;
}

static int pending(void)
{
	if(devfd < 0) return 0;
	return ser_pending(devfd);
}

static int getevent(device_event *ev)
{
	proc_packets();

	if(!QEMPTY(evq)) {
		*ev = evq[evq_ridx];
		evq_ridx = QNEXT(evq_ridx);
		return 1;
	}
	return 0;
}

static int proc_packets(void)
{
	int i, count = 0;
	device_event *ev;
	char buf[128];
	unsigned int st, delta, prev;

	if(devfd < 0) return -1;

	while(ser_getline(devfd, buf, sizeof buf)) {

		switch(buf[0]) {
		case 'd':
			ev = evq + evq_widx;
			if(parse_motion(ev->motion.motion, buf + 1) == -1) {
				break;
			}
			ev->type = DEV_EV_MOTION;
			enqueue_event(ev);
			++count;
			break;

		case 'k':
			if(parse_keystate(&st, buf + 1) == -1) {
				break;
			}

			delta = st ^ bnstate;
			prev = bnstate;
			bnstate = st;

			for(i=0; i<32; i++) {
				if(delta & 1) {
					ev = evq + evq_widx;
					ev->type = DEV_EV_BUTTON;
					ev->button.id = i;
					ev->button.pressed = st & 1;
					ev->button.state = prev ^ (1 << i);
					enqueue_event(ev);
					++count;
				}
				st >>= 1;
				delta >>= 1;
			}
			break;
		}
	}
	return count;
}

static void enqueue_event(device_event *ev)
{
	if(ev != evq + evq_widx) {
		evq[evq_widx] = *ev;
	}

	evq_widx = QNEXT(evq_widx);
	if(evq_widx == evq_ridx) {
		/* overflowed, drop the oldest event */
		evq_ridx = QNEXT(evq_ridx);
	}
}

static int parse_motion(int *motv, const char *data)
{
	int i;
	long val;

	for(i=0; i<6; i++) {
		if(!data[0] || !data[1] || !data[2] || !data[3]) {
			return -1;
		}
		val = ((((long)data[0] & 0xf) << 12) |
			(((long)data[1] & 0xf) << 8) |
			(((long)data[2] & 0xf) << 4) |
			((long)data[3] & 0xf)) - 32768;
		data += 4;
		*motv++ = (int)val;
	}
	return 0;
}

static int parse_keystate(unsigned int *stptr, const char *data)
{
	int i, bit = 0;
	unsigned int st = 0;

	for(i=0; i<3; i++) {
		if(!data) return -1;
		st |= ((unsigned int)*data++ & 0xf) << bit;
		bit += 4;
	}
	*stptr = st;
	return 0;
}
